#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-02-11
# @Filename: database.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import datetime
import os
import pathlib

from typing import TYPE_CHECKING

import polars

from too import log
from too.datamodel import too_dtypes, too_metadata_columns
from too.tools import deduplicate_too_targets, read_too_file
from too.validate import validate_too_targets


if TYPE_CHECKING:
    from sdssdb.connection import PeeweeDatabaseConnection


__all__ = ["connect_to_database", "get_database_uri", "load_too_targets"]

DEFAULT_USER: str = "sdss"
DEFAULT_HOST = "localhost"


def connect_to_database(
    dbname: str,
    host: str | None = None,
    port: int | None = None,
    user: str | None = None,
    password: str | None = None,
):
    """Connects the ``sdssdb`` ``sdss5db`` models to the database."""

    from sdssdb.peewee.sdss5db import catalogdb

    if port is None:
        port = int(os.environ.get("PGPORT", 5432))

    catalogdb.database.connect(
        dbname,
        host=host,
        port=port,
        user=user,
        password=password,
    )

    if not catalogdb.database.connected:
        raise RuntimeError("Could not connect to the database.")

    return catalogdb.database


def get_database_uri(
    dbname: str,
    host: str | None = None,
    port: int | None = None,
    user: str | None = None,
    password: str | None = None,
):
    """Returns the URI to the database."""

    user = user or str(os.environ.get("PGUSER", DEFAULT_USER))
    host = host or str(os.environ.get("PGHOST", DEFAULT_HOST))
    port = port or int(os.environ.get("PGPORT", 5432))

    if user is None and password is None:  # pragma: no cover
        # This should never happen. user is always set.
        auth: str = ""
    elif password is None:
        auth: str = f"{user}@"
    else:
        auth: str = f"{user}:{password}@"

    host_port: str = f"{host or ''}" if port is None else f"{host or ''}:{port}"

    return f"postgresql://{auth}{host_port}/{dbname}"


def database_uri_from_connection(database: PeeweeDatabaseConnection):
    """Returns a database URI from a connection."""

    return get_database_uri(database.dbname, **database.connect_params)


def load_too_targets(
    targets: polars.DataFrame | str | pathlib.Path,
    database: PeeweeDatabaseConnection,
    validate_magnitude_limits: bool = False,
    update_existing: bool = False,
):
    """Loads a list of ToO targets into the database."""

    assert database.connected, "Database is not connected."

    now = datetime.datetime.now(datetime.UTC)

    targets = read_too_file(targets, cast=True)
    targets = deduplicate_too_targets(targets)
    targets = validate_too_targets(
        targets,
        database,
        bright_limit_checks=validate_magnitude_limits,
    )

    # We always want to set the optical_prov column to too.
    targets = targets.with_columns(optical_prov=polars.lit("too"))

    too_columns = ["too_id"]
    too_columns += [col for col in too_dtypes if col not in too_metadata_columns]

    too_target_new = targets.select(too_columns)
    too_metadata_new = targets.select(too_metadata_columns)

    database_uri = get_database_uri(database.dbname, **database.connect_params)

    db_targets = polars.read_database_uri(
        "SELECT * from catalogdb.too_target ORDER BY too_id",
        database_uri,
        engine="adbc",
    )

    new_in_db = too_target_new.filter(polars.col.too_id.is_in(db_targets["too_id"]))
    if not new_in_db.equals(db_targets):
        log.warning(
            "Some ToO targets are already in too_target. Only changes "
            "to metadata columns will be applied for those."
        )

    new_targets = too_target_new.filter(~polars.col.too_id.is_in(db_targets["too_id"]))
    if len(new_targets) == 0:
        log.info("No new ToO targets to add.")
        if update_existing is False:
            return polars.DataFrame(schema=too_dtypes)
    else:
        log.info(f"Loading {len(new_targets)} new ToO(s) into catalogdb.too_target.")

        # Add current date-time.
        new_targets = new_targets.with_columns(added_date=polars.lit(now))

        new_targets.write_database(
            "catalogdb.too_target",
            database_uri,
            if_table_exists="append",
            engine="adbc",
        )

        if not update_existing:
            too_metadata_new = too_metadata_new.filter(
                polars.col.too_id.is_in(new_targets["too_id"])
            )
            update_existing = True

    if update_existing:
        log.info("Updating ToO entries in catalogdb.too_metadata.")

        # First get the current metadata.
        too_metadata = polars.read_database_uri(
            "SELECT * FROM catalogdb.too_metadata ORDER BY too_id",
            database_uri,
            engine="adbc",
        ).with_columns(
            last_modified_date=polars.col.last_modified_date.cast(
                polars.Datetime(time_zone="UTC")
            )
        )

        # Update the metadata dataframe.
        too_metadata_new = too_metadata_new.with_columns(
            last_modified_date=polars.lit(now)
        )
        too_metadata = too_metadata.update(
            too_metadata_new,
            on="too_id",
            how="full",
            include_nulls=True,
        )

        # Replace the table (this is faster than atomic updates).
        too_metadata.write_database(
            "catalogdb.too_metadata",
            database_uri,
            if_table_exists="replace",
            engine="adbc",
        )

        # Recreate PK nand FK constraints.
        database.execute_sql(
            "ALTER TABLE catalogdb.too_metadata ADD PRIMARY KEY (too_id);"
        )
        database.execute_sql(
            "ALTER TABLE catalogdb.too_metadata ADD FOREIGN KEY (too_id) "
            "REFERENCES catalogdb.too_target (too_id) ON DELETE CASCADE;"
        )

    log.debug("Running VACUUM ANALYZE on catalogdb.too_target")
    database.execute_sql("VACUUM ANALYZE catalogdb.too_target;")
    database.execute_sql("VACUUM ANALYZE catalogdb.too_metadata;")

    return new_targets



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-02-08
# @Filename: datamodel.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

# ruff: noqa: E501

from __future__ import annotations

import polars
import polars.type_aliases


__all__ = ["too_dtypes"]


too_dtypes: polars.type_aliases.SchemaDefinition = {
    "too_id": polars.Int64,
    "program": polars.String,
    "fiber_type": polars.String,
    "catalogid": polars.Int64,
    "sdss_id": polars.Int64,
    "gaia_dr3_source_id": polars.Int64,
    "twomass_pts_key": polars.Int32,
    "sky_brightness_mode": polars.String,
    "ra": polars.Float64,
    "dec": polars.Float64,
    "pmra": polars.Float32,
    "pmdec": polars.Float32,
    "epoch": polars.Float32,
    "parallax": polars.Float32,
    "lambda_eff": polars.Float32,
    "u_mag": polars.Float32,
    "g_mag": polars.Float32,
    "r_mag": polars.Float32,
    "i_mag": polars.Float32,
    "z_mag": polars.Float32,
    "optical_prov": polars.String,
    "gaia_bp_mag": polars.Float32,
    "gaia_rp_mag": polars.Float32,
    "gaia_g_mag": polars.Float32,
    "h_mag": polars.Float32,
    "delta_ra": polars.Float32,
    "delta_dec": polars.Float32,
    "can_offset": polars.Boolean,
    "inertial": polars.Boolean,
    "n_exposures": polars.Int16,
    "priority": polars.Int16,
    "active": polars.Boolean,
    "observe_from_mjd": polars.Int32,
    "observe_until_mjd": polars.Int32,
    "observed": polars.Boolean,
    "added_on": polars.Date,
}

too_metadata_columns = [
    "too_id",
    "sky_brightness_mode",
    "lambda_eff",
    "u_mag",
    "g_mag",
    "r_mag",
    "i_mag",
    "z_mag",
    "optical_prov",
    "gaia_bp_mag",
    "gaia_rp_mag",
    "gaia_g_mag",
    "h_mag",
    "delta_ra",
    "delta_dec",
    "can_offset",
    "inertial",
    "n_exposures",
    "priority",
    "active",
    "observe_from_mjd",
    "observe_until_mjd",
    "observed",
    "added_on",
]

column_descriptions = {
    "too_id": "Unique identifier for the ToO target. [required]",
    "program": "The time-domain program that provided this target.",
    "fiber_type": "Type of fiber to be used to observe the target (`APOGEE` or `BOSS`). [required]",
    "catalogid": "catalogid for this target, if already matched.",
    "sdss_id": "sdss_id for this target, if already matched.",
    "gaia_dr3_source_id": "Unique identifier of the target in the Gaia DR3 catalog (`source_id` column).",
    "twomass_pts_key": "Unique identifier of the target in the 2MASS catalog (`pts_key` column).",
    "sky_brightness_mode": "The sky brightness mode. Either `bright` or `dark`. [required]",
    "ra": "The Right Ascension of the target in ICRS coordinates as decimal degrees. [required]",
    "dec": "The declination of the target in ICRS coordinates as decimal degrees. [required]",
    "pmra": "The proper motion in RA in `mas/yr` as a true angle (the `cos(dec)` factor must have been applied).",
    "pmdec": "The proper motion in Dec in `mas/yr` as a true angle.",
    "epoch": "The epoch of the `ra`/`dec` coordinates. Required when `pmra`/`pmdec` are defined.",
    "parallax": "The parallax in `mas`.",
    "lambda_eff": "The effective wavelength to calculate the atmospheric refraction correction for the target.",
    "u_mag": "Sloan `u'` magnitude of the target.",
    "g_mag": "Sloan `g'` magnitude of the target.",
    "r_mag": "Sloan `r'` magnitude of the target.",
    "i_mag": "Sloan `i'` magnitude of the target.",
    "z_mag": "Sloan `z'` magnitude of the target.",
    "optical_prov": "Source of the optical magnitudes.",
    "gaia_bp_mag": "Gaia `BP` magnitude of the target.",
    "gaia_rp_mag": "Gaia `RP` magnitude of the target.",
    "gaia_g_mag": "Gaia `G` magnitude of the target.",
    "h_mag": "`H`-band magnitude of the target.",
    "delta_ra": "Fixed RA offset as a true angle, in arcsec.",
    "delta_dec": "Fixed Dec offset, in arcsec.",
    "can_offset": "`true` if the ToO is allowed to be offset. Offseting will only occur if target violates magnitude limits. [required]",
    "inertial": "If `true`, the proper motions for this target will be ignored (equivalent to `pmra = pmdec = 0.0`).",
    "n_exposures": "The minimum number of exposures required for the ToO to be complete and not assigned anymore. [required]",
    "priority": "The relative prioriry of this target (10: highest, 1: lowest, 0: the target will be ignored)",
    "active": "`true` if the target is active and should be assigned to configurations if possible. [required]",
    "observe_from_mjd": "MJD from which the target is considered observable. Default to the current date.",
    "observe_until_mjd": "MJD at which the target should automatically be consider inactive. If empty, the target never expires.",
    "observed": "`true` if the target has been fully observed and should not be assigned again. [required]",
    "added_on": "Date when the target was last added or modified, with the format YYYY-MM-DD. [required]",
}

mag_columns = [
    "u_mag",
    "g_mag",
    "r_mag",
    "i_mag",
    "z_mag",
    "gaia_bp_mag",
    "gaia_rp_mag",
    "gaia_g_mag",
    "h_mag",
]

fiber_type_values = ["APOGEE", "BOSS"]


def datamodel_to_markdown():  # pragma: no cover
    """Prints the datamodel as a markdown table."""

    df = polars.DataFrame(
        {
            "Column": list(too_dtypes.keys()),
            "Type": [v.__name__.lower() for v in too_dtypes.values()],
            "Description": [column_descriptions.get(k, "") for k in too_dtypes.keys()],
        }
    )

    with polars.Config(
        tbl_formatting="ASCII_MARKDOWN",
        tbl_hide_column_data_types=True,
        tbl_hide_dataframe_shape=True,
        tbl_rows=-1,
        tbl_width_chars=10000,
        fmt_str_lengths=10000,
    ):
        print(df)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-02-17
# @Filename: __main__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import enum
import pathlib

from typing import Annotated

import typer
from typer.core import TyperGroup


class NaturalOrderGroup(TyperGroup):
    """A Typer group that lists commands in order of definition."""

    def list_commands(self, ctx):
        return self.commands.keys()


too_cli = typer.Typer(
    cls=NaturalOrderGroup,
    rich_markup_mode="rich",
    context_settings={"obj": {}},
    no_args_is_help=True,
    help="Command line interface for ToOs.",
)

dbname = typer.Option(
    help="The name of the database to connect to.",
    rich_help_panel="Options for connecting to the database server",
    envvar="TOO_DBNAME",
)
host = typer.Option(
    help="The host of the database server.",
    rich_help_panel="Options for connecting to the database server",
    envvar="TOO_DBHOST",
)
port = typer.Option(
    help="The port of the database server.",
    rich_help_panel="Options for connecting to the database server",
    envvar="TOO_DBPORT",
)
user = typer.Option(
    help="The user to connect to the database.",
    rich_help_panel="Options for connecting to the database server",
    envvar="TOO_DBUSER",
)


class Observatories(enum.Enum):
    APO = "APO"
    LCO = "LCO"


@too_cli.command()
def process(
    files: Annotated[
        list[pathlib.Path] | None,
        typer.Argument(
            exists=True,
            dir_okay=True,
            envvar="TOO_TARGET_FILES",
            help="The list of files to process. If a directory is passed, all the "
            "files in the directory will be processed.",
        ),
    ] = None,
    ignore_invalid: Annotated[
        bool,
        typer.Option(
            "--ignore-invalid",
            help="Exclude target files that cannot be read or cast.",
        ),
    ] = False,
    verbose: Annotated[
        bool,
        typer.Option("-v", "--verbose", help="Prints more information."),
    ] = False,
    write_log: Annotated[
        pathlib.Path | None,
        typer.Option(help="Path where to write the log file."),
    ] = None,
    dbname: Annotated[str, dbname] = "sdss5db",
    host: Annotated[str, host] = "localhost",
    port: Annotated[int | None, port] = None,
    user: Annotated[str, user] = "sdss",
    cross_match: Annotated[
        bool,
        typer.Option(
            help="Cross-match the input files."
            "If --no-cross-match is passed, the carton is not run.",
            rich_help_panel="Options for loading ToOs and running the too carton.",
        ),
    ] = True,
    run_carton: Annotated[
        bool,
        typer.Option(
            help="Run the carton after cross-matching.",
            rich_help_panel="Options for loading ToOs and running the too carton.",
        ),
    ] = True,
):
    """Ingest and process new lists of ToOs."""

    import polars

    from too import (
        connect_to_database,
        load_too_targets,
        log,
        read_too_file,
        run_too_carton,
        too_dtypes,
        update_sdss_id_tables,
        xmatch_too_targets,
    )

    if verbose:
        log.sh.setLevel("DEBUG")

    if write_log:
        log.start_file_logger(str(write_log), rotating=False)

    database = connect_to_database(
        dbname,
        host=host,
        port=port,
        user=user,
    )

    if files is not None and len(files) > 0:
        process_files: list[pathlib.Path] = []
        for file in files:
            if file.is_dir():
                process_files.extend(file.glob("*.csv"))
                process_files.extend(file.glob("*.parquet"))
            else:
                process_files.append(file)

        log.debug(f"Reading {len(process_files)} input file(s).")
        targets = polars.DataFrame({}, schema=too_dtypes)
        for file in process_files:
            try:
                new_targets = read_too_file(file, cast=True)
            except Exception as ee:
                log.error(f"Failed to read file {file}: {ee}")
                if ignore_invalid:
                    continue
                else:
                    raise
            else:
                targets = targets.vstack(new_targets)
        targets = targets.sort(["added_on", "too_id"])

        log.info("Loading targets into the database.")
        load_too_targets(
            targets,
            database,
            update_existing=True,
            validate_magnitude_limits=True,
        )

    else:
        log.warning("No new files to process.")

    if cross_match:
        log.info("Cross-matching targets.")
        xmatch_too_targets(database)

        if run_carton:
            log.info("Running the ToO carton.")
            run_too_carton()
            update_sdss_id_tables(database)


@too_cli.command(name="dump-targets")
def dump_targets(
    file: Annotated[str, typer.Argument(help="The file to dump the ToO targets into.")],
    observatory: Annotated[
        Observatories,
        typer.Option(help="The observatory for which to dump the ToO targets."),
    ],
    dbname: Annotated[str, dbname] = "sdss5db",
    host: Annotated[str, host] = "localhost",
    port: Annotated[int | None, port] = None,
    user: Annotated[str, user] = "sdss",
):
    """Dumps the current ToO targets into a Parquet file."""

    from too import connect_to_database, dump_targets_to_parquet

    database = connect_to_database(dbname, host=host, port=port, user=user)
    dump_targets_to_parquet(observatory.value.upper(), file, database=database)


@too_cli.command(name="dump-sdss-id")
def dump_sdss_id(
    last_updated: Annotated[
        str,
        typer.Argument(help="The last_updated value to use to search for new targets."),
    ],
    root: Annotated[
        str,
        typer.Option(help="The root directory where to save the CSV files."),
    ] = ".",
    dbname: Annotated[str, dbname] = "sdss5db",
    host: Annotated[str, host] = "localhost",
    port: Annotated[int | None, port] = None,
    user: Annotated[str, user] = "sdss",
):
    """Dumps the ``sdss_id_flat`` and ``sdss_id_stacked`` tables for a given date.

    Searches the ``sandbox.sdss_id_stacked`` table for targets matching
    ``last_updated`` and uses that to determine the ``sdss_id_flat`` associated rows.
    The tables are saved to the current directory as CSV files.

    """

    from too import connect_to_database, dump_sdss_id_tables

    database = connect_to_database(
        dbname,
        host=host,
        port=port,
        user=user,
    )

    dump_sdss_id_tables(last_updated, database, root=root, overwrite=True)


@too_cli.command()
def validate(
    file: Annotated[str, typer.Argument(help="The file to validate.")],
    cast: Annotated[
        bool,
        typer.Option(help="Cast the columns to the datamodel types."),
    ],
):
    """Validates a ToO file."""

    import sys

    from too import log, read_too_file, validate_too_targets
    from too.exceptions import ValidationError

    df = read_too_file(file, cast=cast)

    try:
        validate_too_targets(df)
    except ValidationError as ee:
        log.error(f"Validation failed with error: {ee}")
        sys.exit(1)
    except Exception as ee:
        log.error(f"Unexpected error: {ee}")
        sys.exit(1)


if __name__ == "__main__":
    too_cli()
